@@ -418,15 +418,17 @@ class Input(object):
         return v, der_v
 
 
-def apply_start_values(fmu, model_description, start_values, apply_default_start_values=False, validate=True):
+def apply_start_values(fmu, model_description, start_values, settable=None):
     """ Set start values to an FMU instance
 
     Parameters:
-        fmu                     the FMU instance
-        model_description       the ModelDescription instance
-        start_values            dictionary of variable_name -> start_value pairs
-        apply_default_values    apply the start values from the model description
-        validate                check if the start values can be set
+        fmu                the FMU instance
+        model_description  the ModelDescription instance
+        start_values       dictionary of variable_name -> start_value pairs
+        settable           callback f(variable) -> bool that indicates if a variable can be set
+
+    returns:
+        a dictionary with the start values that have not been set
     """
 
     start_values = start_values.copy()
@@ -435,16 +437,18 @@ def apply_start_values(fmu, model_description, start_values, apply_default_start
 
     for variable in model_description.modelVariables:
 
-        if variable.name in start_values:
-            value = start_values.pop(variable.name)
-        elif apply_default_start_values and variable.start is not None and variable.initial in ['approx', 'exact']:
-            value = variable.start
-        else:
+        if variable.name not in start_values:
             continue
 
+        if settable is not None and not settable(variable):
+            continue
+
+        value = start_values.pop(variable.name)
+
         if type(value) is tuple:
             if len(value) != 2:
-                raise Exception('The start value for variable %s must be a scalar value or a tuple (<value>, {<unit>|<display_unit>}) but was "%s".' % (variable.name, value))
+                raise Exception(f'The start value for variable {variable.name} must be a scalar value or a'
+                                f' tuple (<value>, {{<unit>|<display_unit>}}) but was "{value}".')
             value, unit = value
         else:
             unit = None
@@ -459,22 +463,15 @@ def apply_start_values(fmu, model_description, start_values, apply_default_start
             elif variable.declaredType is not None:
                 base_unit = variable.declaredType.unit
             else:
-                raise Exception('Variable %s has no unit but the unit "%s" was specified for its start value.' % (variable.name, unit))
+                raise Exception(f'Variable {variable.name} has no unit but the unit "{unit}"'
+                                ' was specified for its start value.')
 
             if unit not in unit_definitions[base_unit]:
-                raise Exception('The unit "%s" of the start value for variable %s is not defined.' % (unit, variable.name))
+                raise Exception(f'The unit "{unit}" of the start value for variable {variable.name} is not defined.')
 
             display_unit = unit_definitions[base_unit][unit]
             value = (value - display_unit.offset) / display_unit.factor
 
-        if validate:  # check if start value can be set
-
-            if variable.variability == 'constant':
-                raise Exception(f'Variable "{variable.name}" is a constant and cannot be set.')
-
-            if variable.initial not in ['approx', 'exact']:
-                raise Exception(f'Variable "{variable.name}" has inital={variable.initial} and cannot be set.')
-
         vr = variable.valueReference
 
         # get the setter function
@@ -490,8 +487,8 @@ def apply_start_values(fmu, model_description, start_values, apply_default_start
         if variable.type == 'Boolean':
             if isinstance(value, str):
                 if value.lower() not in ['true', 'false']:
-                    raise Exception('The start value "%s" for variable "%s" could not be converted to Boolean' %
-                                    (value, variable.name))
+                    raise Exception(f'The start value "{value}" for variable "{variable.name}"'
+                                    ' could not be converted to Boolean')
                 else:
                     value = value.lower() == 'true'
 
@@ -501,16 +498,14 @@ def apply_start_values(fmu, model_description, start_values, apply_default_start
                 value = value.split()
             value = list(map(lambda e: variable._python_type(e), value))
             if len(value) != np.prod(variable.shape):
-                raise ArgumentError('The start value for variable "%s" must have %d elements but has %d.' %
-                                    (variable.name, np.prod(variable.shape), len(value)))
+                raise ArgumentError(f'The start value for variable "{variable.name}" must have'
+                                    f' {np.prod(variable.shape)} elements but has {len(value)}.')
         else:
             value = [variable._python_type(value)]
 
         setter([vr], value)
 
-    if len(start_values) > 0:
-        raise Exception("The start values for the following variables could not be set because they don't exist: " +
-                        ', '.join(start_values.keys()))
+    return start_values
 
 
 class ForwardEuler(object):
@@ -571,7 +566,7 @@ def simulate_fmu(filename,
                  validate: bool = True,
                  start_time: Union[float, str] = None,
                  stop_time: Union[float, str] = None,
-                 solver: str ='CVode',
+                 solver: str = 'CVode',
                  step_size: Union[float, str] = None,
                  relative_tolerance: Union[float, str] = None,
                  output_interval: Union[float, str] = None,
@@ -592,7 +587,10 @@ def simulate_fmu(filename,
                  set_input_derivatives: bool = False,
                  remote_platform: str = 'auto',
                  early_return_allowed: bool = False,
-                 use_event_mode: bool = False) -> SimulationResult:
+                 use_event_mode: bool = False,
+                 initialize: bool = True,
+                 terminate: bool = True,
+                 fmu_state: Union[bytes, c_void_p] = None) -> SimulationResult:
     """ Simulate an FMU
 
     Parameters:
@@ -623,13 +621,16 @@ def simulate_fmu(filename,
                                is not supported, None: no remoting; experimental)
         early_return_allowed   allow early return in FMI 3.0 Co-Simulation
         use_event_mode         use event mode in FMI 3.0 Co-Simulation if the FMU supports it
+        initialize             initialize the FMU
+        terminate              terminate the FMU
+        fmu_state              the FMU state or serialized FMU state to initialize the FMU
     Returns:
         result                 a structured numpy array that contains the result
     """
 
     from fmpy import supported_platforms
     from fmpy.model_description import read_model_description
-    from fmpy.util import has_wine64, has_wsl, can_simulate
+    from fmpy.util import can_simulate
 
     platforms = supported_platforms(filename)
 
@@ -655,6 +656,12 @@ def simulate_fmu(filename,
     if fmi_type not in ['ModelExchange', 'CoSimulation']:
         raise Exception('fmi_type must be one of "ModelExchange" or "CoSimulation"')
 
+    if initialize is False:
+        if fmi_type != 'CoSimulation':
+            raise Exception("If initialize is False, the interface type must be 'CoSimulation'.")
+        if fmu_instance is None and fmu_state is None:
+            raise Exception("If initialize is False, fmu_instance or fmu_state must be provided.")
+
     experiment = model_description.defaultExperiment
 
     if start_time is None:
@@ -711,11 +718,23 @@ def simulate_fmu(filename,
     else:
         fmu = fmu_instance
 
+    if fmu_state is not None:
+        if model_description.fmiVersion == '2.0' or model_description.fmiVersion.startswith('3.0'):
+            if isinstance(fmu_state, bytes):
+                fmu_state = fmu.deserializeFMUState(fmu_state)
+                fmu.setFMUState(fmu_state)
+                fmu.freeFMUState(fmu_state)
+            else:
+                fmu.setFMUState(fmu_state)
+        else:
+            raise Exception(f"Setting the FMU state is not supported for FMI version {model_description.fmiVersion}.")
+        initialize = False
+
     # simulate_fmu the FMU
     if fmi_type == 'ModelExchange':
         result = simulateME(model_description, fmu, start_time, stop_time, solver, step_size, relative_tolerance, start_values, apply_default_start_values, input, output, output_interval, record_events, timeout, step_finished, validate)
     elif fmi_type == 'CoSimulation':
-        result = simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance, start_values, apply_default_start_values, input, output, output_interval, timeout, step_finished, set_input_derivatives, use_event_mode, early_return_allowed, validate)
+        result = simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance, start_values, apply_default_start_values, input, output, output_interval, timeout, step_finished, set_input_derivatives, use_event_mode, early_return_allowed, validate, initialize, terminate)
 
     if fmu_instance is None:
         fmu.freeInstance()
@@ -743,21 +762,18 @@ def instantiate_fmu(unzipdir, model_description, fmi_type=None, visible=False, d
     if library_path:
         fmu_args['libraryPath'] = library_path
 
-    if logger is None:
-        logger = printLogMessage
-
     is_fmi1 = model_description.fmiVersion == '1.0'
     is_fmi2 = model_description.fmiVersion == '2.0'
 
     if is_fmi1:
         callbacks = fmi1CallbackFunctions()
-        callbacks.logger         = fmi1CallbackLoggerTYPE(logger)
+        callbacks.logger         = fmi1CallbackLoggerTYPE(printLogMessage if logger is None else logger)
         callbacks.allocateMemory = fmi1CallbackAllocateMemoryTYPE(allocateMemory)
         callbacks.freeMemory     = fmi1CallbackFreeMemoryTYPE(freeMemory)
         callbacks.stepFinished   = None
     elif is_fmi2:
         callbacks = fmi2CallbackFunctions()
-        callbacks.logger         = fmi2CallbackLoggerTYPE(logger)
+        callbacks.logger         = fmi2CallbackLoggerTYPE(printLogMessage if logger is None else logger)
         callbacks.allocateMemory = fmi2CallbackAllocateMemoryTYPE(allocateMemory)
         callbacks.freeMemory     = fmi2CallbackFreeMemoryTYPE(freeMemory)
     else:
@@ -783,7 +799,8 @@ def instantiate_fmu(unzipdir, model_description, fmi_type=None, visible=False, d
             fmu.instantiate(visible=visible, callbacks=callbacks, loggingOn=debug_logging)
         else:
             fmu = fmi3.FMU3Slave(**fmu_args)
-            fmu.instantiate(visible=visible, loggingOn=debug_logging, eventModeUsed=event_mode_used, earlyReturnAllowed=early_return_allowed)
+            fmu.instantiate(visible=visible, loggingOn=debug_logging, eventModeUsed=event_mode_used,
+                            earlyReturnAllowed=early_return_allowed, logMessage=logger)
 
     elif fmi_type in [None, 'ModelExchange'] and model_description.modelExchange is not None:
 
@@ -797,13 +814,13 @@ def instantiate_fmu(unzipdir, model_description, fmi_type=None, visible=False, d
             fmu.instantiate(visible=visible, callbacks=callbacks, loggingOn=debug_logging)
         else:
             fmu = fmi3.FMU3Model(**fmu_args)
-            fmu.instantiate(visible=visible, loggingOn=debug_logging)
+            fmu.instantiate(visible=visible, loggingOn=debug_logging, logMessage=logger)
 
     elif fmi_type in [None, 'ScheduledExecution'] and model_description.scheduledExecution is not None:
 
         fmu_args['modelIdentifier'] = model_description.scheduledExecution.modelIdentifier
         fmu = fmi3.FMU3ScheduledExecution(**fmu_args)
-        fmu.instantiate(visible=visible, loggingOn=debug_logging)
+        fmu.instantiate(visible=visible, loggingOn=debug_logging, logMessage=logger)
 
     else:
 
@@ -812,6 +829,20 @@ def instantiate_fmu(unzipdir, model_description, fmi_type=None, visible=False, d
     return fmu
 
 
+def has_start_value(variable):
+    return variable.start is not None
+
+
+def settable_in_instantiated(variable):
+    return variable.variability != 'constant' and variable.initial in {'approx', 'exact'}
+
+
+def settable_in_initialization_mode(variable):
+    return variable.causality == 'input' \
+           or (variable.causality != 'parameter' and variable.variability == 'tunable') \
+           or (variable.variability != 'constant' and variable.initial == 'exact')
+
+
 def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_size, relative_tolerance, start_values, apply_default_start_values, input_signals, output, output_interval, record_events, timeout, step_finished, validate):
 
     if relative_tolerance is None:
@@ -846,59 +877,69 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
 
     input = Input(fmu, model_description, input_signals)
 
-    apply_start_values(fmu, model_description, start_values, apply_default_start_values, validate)
-
     # initialize
     if is_fmi1:
+        start_values = apply_start_values(fmu, model_description, start_values, settable=has_start_value)
 
         input.apply(time)
 
         (iterationConverged,
          stateValueReferencesChanged,
          stateValuesChanged,
-         terminateSimulation,
+         terminate_simulation,
          nextEventTimeDefined,
          nextEventTime) = fmu.initialize()
 
-        if terminateSimulation:
+        if terminate_simulation:
             raise Exception('Model requested termination during initial event update.')
 
     elif is_fmi2:
 
+        start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_instantiated)
+
         fmu.enterInitializationMode()
+
+        start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_initialization_mode)
+
         input.apply(time)
+
         fmu.exitInitializationMode()
 
         newDiscreteStatesNeeded = True
-        terminateSimulation = False
+        terminate_simulation = False
 
-        while newDiscreteStatesNeeded and not terminateSimulation:
-            # update discrete states
+        while newDiscreteStatesNeeded and not terminate_simulation:
             (newDiscreteStatesNeeded,
-             terminateSimulation,
+             terminate_simulation,
              nominalsOfContinuousStatesChanged,
              valuesOfContinuousStatesChanged,
              nextEventTimeDefined,
              nextEventTime) = fmu.newDiscreteStates()
 
-        if terminateSimulation:
+        if terminate_simulation:
             raise Exception('Model requested termination during initial event update.')
 
         fmu.enterContinuousTimeMode()
 
     elif is_fmi3:
 
+        start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_instantiated)
+
         fmu.enterInitializationMode(startTime=start_time, stopTime=stop_time)
+
         input.apply(time)
+
+        start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_initialization_mode)
+
         fmu.exitInitializationMode()
 
         discreteStatesNeedUpdate = True
-        terminateSimulation = False
+        terminate_simulation = False
 
-        while discreteStatesNeedUpdate and not terminateSimulation:
+        while discreteStatesNeedUpdate and not terminate_simulation:
             # update discrete states
             (discreteStatesNeedUpdate,
-             terminateSimulation,
+             terminate_simulation,
              nominalsOfContinuousStatesChanged,
              valuesOfContinuousStatesChanged,
              nextEventTimeDefined,
@@ -906,6 +947,10 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
 
         fmu.enterContinuousTimeMode()
 
+    if validate and len(start_values) > 0:
+        raise Exception("The start values for the following variables could not be set: " +
+                        ', '.join(start_values.keys()))
+
     # common solver constructor arguments
     solver_args = {
         'nx': model_description.numberOfContinuousStates,
@@ -1018,15 +1063,15 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
                 iterationConverged = False
 
                 # update discrete states
-                while not iterationConverged and not terminateSimulation:
+                while not iterationConverged and not terminate_simulation:
                     (iterationConverged,
                      stateValueReferencesChanged,
                      stateValuesChanged,
-                     terminateSimulation,
+                     terminate_simulation,
                      nextEventTimeDefined,
                      nextEventTime) = fmu.eventUpdate()
 
-                if terminateSimulation:
+                if terminate_simulation:
                     break
 
             elif is_fmi2:
@@ -1039,15 +1084,15 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
                 newDiscreteStatesNeeded = True
 
                 # update discrete states
-                while newDiscreteStatesNeeded and not terminateSimulation:
+                while newDiscreteStatesNeeded and not terminate_simulation:
                     (newDiscreteStatesNeeded,
-                     terminateSimulation,
+                     terminate_simulation,
                      nominalsOfContinuousStatesChanged,
                      valuesOfContinuousStatesChanged,
                      nextEventTimeDefined,
                      nextEventTime) = fmu.newDiscreteStates()
 
-                if terminateSimulation:
+                if terminate_simulation:
                     break
 
                 fmu.enterContinuousTimeMode()
@@ -1062,15 +1107,15 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
                 newDiscreteStatesNeeded = True
 
                 # update discrete states
-                while newDiscreteStatesNeeded and not terminateSimulation:
+                while newDiscreteStatesNeeded and not terminate_simulation:
                     (newDiscreteStatesNeeded,
-                     terminateSimulation,
+                     terminate_simulation,
                      nominalsOfContinuousStatesChanged,
                      valuesOfContinuousStatesChanged,
                      nextEventTimeDefined,
                      nextEventTime) = fmu.updateDiscreteStates()
 
-                if terminateSimulation:
+                if terminate_simulation:
                     break
 
                 fmu.enterContinuousTimeMode()
@@ -1095,7 +1140,7 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
     return recorder.result()
 
 
-def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance, start_values, apply_default_start_values, input_signals, output, output_interval, timeout, step_finished, set_input_derivatives, use_event_mode, early_return_allowed, validate):
+def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance, start_values, apply_default_start_values, input_signals, output, output_interval, timeout, step_finished, set_input_derivatives, use_event_mode, early_return_allowed, validate, initialize, terminate):
 
     if set_input_derivatives and not model_description.coSimulation.canInterpolateInputs:
         raise Exception("Parameter set_input_derivatives is True but the FMU cannot interpolate inputs.")
@@ -1108,47 +1153,54 @@ def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance
     is_fmi1 = model_description.fmiVersion == '1.0'
     is_fmi2 = model_description.fmiVersion == '2.0'
 
-    if is_fmi2:
-        fmu.setupExperiment(tolerance=relative_tolerance, startTime=start_time, stopTime=stop_time)
-
     input = Input(fmu=fmu, modelDescription=model_description, signals=input_signals, set_input_derivatives=set_input_derivatives)
 
     time = start_time
 
-    apply_start_values(fmu, model_description, start_values, apply_default_start_values, validate)
+    if initialize:
 
-    # initialize the model
-    if is_fmi1:
-        input.apply(time)
-        fmu.initialize(tStart=time, stopTime=stop_time)
-    elif is_fmi2:
-        fmu.enterInitializationMode()
-        input.apply(time)
-        fmu.exitInitializationMode()
-    else:
-        fmu.enterInitializationMode(tolerance=relative_tolerance, startTime=start_time, stopTime=stop_time)
-        input.apply(time)
-        fmu.exitInitializationMode()
+        if is_fmi2:
+            fmu.setupExperiment(tolerance=relative_tolerance, startTime=time, stopTime=stop_time)
 
-        if use_event_mode:
+        # initialize the model
+        if is_fmi1:
+            start_values = apply_start_values(fmu, model_description, start_values, settable=has_start_value)
+            input.apply(time)
+            fmu.initialize(tStart=time, stopTime=stop_time)
+        elif is_fmi2:
+            start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_instantiated)
+            fmu.enterInitializationMode()
+            start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_initialization_mode)
+            input.apply(time)
+            fmu.exitInitializationMode()
+        else:
+            start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_instantiated)
+            fmu.enterInitializationMode(tolerance=relative_tolerance, startTime=time, stopTime=stop_time)
+            start_values = apply_start_values(fmu, model_description, start_values, settable=settable_in_initialization_mode)
+            input.apply(time)
+            fmu.exitInitializationMode()
 
-            update_discrete_states = True
+            if use_event_mode:
 
-            while update_discrete_states:
-                update_discrete_states, terminate, _, _, _, _ = fmu.updateDiscreteStates()
+                update_discrete_states = True
 
-            if terminate:
-                raise Exception("The FMU requested to terminate the simulation during initialization.")
+                while update_discrete_states:
+                    update_discrete_states, terminate_simulation, _, _, _, _ = fmu.updateDiscreteStates()
 
-            fmu.enterStepMode()
+                if terminate_simulation:
+                    raise Exception("The FMU requested to terminate the simulation during initialization.")
+
+                fmu.enterStepMode()
+
+    if validate and len(start_values) > 0:
+        raise Exception("The start values for the following variables could not be set: " +
+                        ', '.join(start_values.keys()))
 
     recorder = Recorder(fmu=fmu, modelDescription=model_description, variableNames=output, interval=output_interval)
 
     n_steps = time / output_interval
 
-    terminate = False
-
-    time = start_time
+    terminate_simulation = False
 
     # simulation loop
     while True:
@@ -1158,12 +1210,12 @@ def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance
         if timeout is not None and (current_time() - sim_start) > timeout:
             break
 
-        if terminate or time >= stop_time:
+        if terminate_simulation or time >= stop_time:
             break
 
         input.apply(time)
 
-        if is_fmi1 or is_fmi2:
+        if is_fmi1:
 
             if time + output_interval <= stop_time:
                 fmu.doStep(currentCommunicationPoint=time, communicationStepSize=output_interval)
@@ -1172,6 +1224,26 @@ def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance
             else:
                 fmu.doStep(currentCommunicationPoint=time, communicationStepSize=stop_time - time)
                 time = stop_time
+
+        elif is_fmi2:
+
+            try:
+                if time + output_interval <= stop_time:
+                    fmu.doStep(currentCommunicationPoint=time, communicationStepSize=output_interval)
+                    n_steps += 1
+                    time = n_steps * output_interval
+                else:
+                    fmu.doStep(currentCommunicationPoint=time, communicationStepSize=stop_time - time)
+                    time = stop_time
+            except FMICallException as e:
+                if e.status == fmi2Discard:
+                    terminated = fmu.getBooleanStatus(fmi2Terminated)
+                    if terminated:
+                        time = fmu.getRealStatus(fmi2LastSuccessfulTime)
+                        recorder.sample(time, force=True)
+                        break
+                else:
+                    raise e
         else:
 
             t_input_event = input.nextEvent(time)
@@ -1182,7 +1254,7 @@ def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance
 
             step_size = t_input_event - time if input_event else t_next - time
 
-            event_encountered, terminate, early_return, last_successful_time = fmu.doStep(currentCommunicationPoint=time, communicationStepSize=step_size)
+            event_encountered, terminate_simulation, early_return, last_successful_time = fmu.doStep(currentCommunicationPoint=time, communicationStepSize=step_size)
 
             if early_return and not early_return_allowed:
                 raise Exception("FMU returned early from doStep() but Early Return is not allowed.")
@@ -1203,17 +1275,19 @@ def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance
 
                 update_discrete_states = True
 
-                while update_discrete_states and not terminate:
-                    update_discrete_states, terminate, _, _, _, _ = fmu.updateDiscreteStates()
+                while update_discrete_states and not terminate_simulation:
+                    update_discrete_states, terminate_simulation, _, _, _, _ = fmu.updateDiscreteStates()
 
-                if terminate:
+                if terminate_simulation:
                     break
 
                 fmu.enterStepMode()
 
         if step_finished is not None and not step_finished(time, recorder):
+            recorder.sample(time, force=True)
             break
 
-    fmu.terminate()
+    if terminate:
+        fmu.terminate()
 
     return recorder.result()
